﻿Imports System.Collections.Generic
Imports System.Linq
Imports System.Text

Public Class Board
    Private board As Integer(,)
    Private m_size As Integer
    Private empty_col, empty_row As Integer

    Public ReadOnly Property Size() As Integer
        Get
            Return m_size
        End Get
    End Property
    Public ReadOnly Property EmptyTileX() As Integer
        Get
            Return empty_col
        End Get
    End Property
    Public ReadOnly Property EmptyTileY() As Integer
        Get
            Return empty_row
        End Get
    End Property
    Public ReadOnly Property Tiles() As Integer(,)
        Get
            Return board
        End Get
    End Property  ' rows x cols  

    Public Property Cells As Integer()
        Get
            Dim arr(m_size * m_size - 1) As Integer
            Dim k As Integer
            For i As Integer = 0 To m_size - 1
                For j As Integer = 0 To m_size - 1
                    arr(k) = Me.board(i, j)
                    k += i
                Next
            Next
            Return arr
        End Get
        Set(ByVal value As Integer())
            Dim k As Integer
            For i As Integer = 0 To m_size - 1
                For j As Integer = 0 To m_size - 1
                    Me.board(i, j) = value(k)
                    k += 1
                Next
            Next
            FindEmpty()
        End Set
    End Property

    Public Sub New(ByVal _board As Integer(,))
        Me.board = _board
        m_size = Math.Min(_board.GetLength(0), _board.GetLength(1))
        FindEmpty()
    End Sub
    Public Sub New(ByVal _cells As Integer())
        If _cells.Length = 9 Then
            m_size = 3
        ElseIf _cells.Length = 16 Then
            m_size = 4
        Else
            Throw New IndexOutOfRangeException("Invalid array size - only 9 or 16 elements acceptable")
            Return
        End If
        ReDim Me.board(m_size - 1, m_size - 1)
        Dim k As Integer
        For i As Integer = 0 To m_size - 1
            For j As Integer = 0 To m_size - 1
                Me.board(i, j) = _cells(k)
                k += 1
            Next
        Next
        FindEmpty()
    End Sub

    Public Function CanGoLeft() As Boolean
        Return (empty_col > 0)
    End Function
    Public Function CanGoRight() As Boolean
        Return (empty_col < (m_size - 1))
    End Function
    Public Function CanGoUp() As Boolean
        Return (empty_row > 0)
    End Function
    Public Function CanGoDown() As Boolean
        Return (empty_row < (m_size - 1))
    End Function

    Public Sub Move(ByVal _move As Direction)
        Select Case _move
            Case Direction.LEFT : MoveLeft()
            Case Direction.RIGHT : MoveRight()
            Case Direction.UP : MoveUp()
            Case Direction.DOWN : MoveDown()
            Case Else
                Return    ' wtf? :)  
        End Select
    End Sub
    Public Sub MoveLeft()
        Dim tmp As Integer = board(empty_row, empty_col)
        board(empty_row, empty_col) = board(empty_row, empty_col - 1)
        board(empty_row, empty_col - 1) = tmp
        empty_col -= 1
    End Sub
    Public Sub MoveRight()
        Dim tmp As Integer = board(empty_row, empty_col)
        board(empty_row, empty_col) = board(empty_row, empty_col + 1)
        board(empty_row, empty_col + 1) = tmp
        empty_col += 1
    End Sub
    Public Sub MoveUp()
        Dim tmp As Integer = board(empty_row, empty_col)
        board(empty_row, empty_col) = board(empty_row - 1, empty_col)
        board(empty_row - 1, empty_col) = tmp
        empty_row -= 1
    End Sub
    Public Sub MoveDown()
        Dim tmp As Integer = board(empty_row, empty_col)
        board(empty_row, empty_col) = board(empty_row + 1, empty_col)
        board(empty_row + 1, empty_col) = tmp
        empty_row += 1
    End Sub

    Public Function GetMoveTile(ByVal _move As Direction) As Integer
        Select Case _move
            Case Direction.LEFT : Return board(empty_row, empty_col - 1)
            Case Direction.RIGHT : Return board(empty_row, empty_col + 1)
            Case Direction.UP : Return board(empty_row - 1, empty_col)
            Case Direction.DOWN : Return board(empty_row + 1, empty_col)
            Case Else
                Return -1   ' wtf? :)  
        End Select
    End Function

    Public Function Successors(ByVal path As List(Of Direction)) As List(Of Direction)
        Dim last_move As Direction = (If(((path IsNot Nothing) AndAlso (path.Count > 0)), path(path.Count - 1), Direction.NONE))
        Dim moves As New List(Of Direction)()
        If CanGoLeft() AndAlso (last_move <> Direction.RIGHT) Then moves.Add(Direction.LEFT)
        If CanGoRight() AndAlso (last_move <> Direction.LEFT) Then moves.Add(Direction.RIGHT)
        If CanGoUp() AndAlso (last_move <> Direction.DOWN) Then moves.Add(Direction.UP)
        If CanGoDown() AndAlso (last_move <> Direction.UP) Then moves.Add(Direction.DOWN)
        Return moves
    End Function
    Public Function MoveCost(ByVal move As Direction) As Integer
        Return 1   ' all valid transformations require only a single move  
    End Function

    Public Function IsGoal() As Boolean
        For i As Integer = 0 To m_size - 1
            For j As Integer = 0 To m_size - 1
                If board(i, j) <> ((i * m_size) + (j + 1)) Then Return False
            Next
        Next
        Return True
    End Function

    Private Sub FindEmpty()
        Dim empty As Integer = m_size * m_size
        For i As Integer = 0 To m_size - 1
            For j As Integer = 0 To m_size - 1
                If board(i, j) = empty Then
                    empty_row = i : empty_col = j
                    i = m_size : j = m_size
                End If
            Next
        Next
    End Sub
End Class
