﻿
Public Class Puzzle
    Private m_size, emptyID, count As Integer
    Private m_board As Board
    Private cells As Integer()

    Public Property BoardSize() As Integer
        Get
            Return m_size
        End Get
        Set(ByVal value As Integer)
            m_size = value
            InitPuzzle()
        End Set
    End Property
    Public ReadOnly Property EmptyValue As Integer
        Get
            Return count
        End Get
    End Property
    Public Property StringConfig As String
        Get
            Return String.Join(",", cells)
        End Get
        Set(ByVal value As String)
            Dim vals As String() = value.Split(","c)
            If vals.Length = 9 Then
                m_size = 3
            ElseIf vals.Length = 16 Then
                m_size = 4
            Else
                Throw New IndexOutOfRangeException("Invalid string - only 9 or 16 elements acceptable")
                Return
            End If
            Dim tempArr(m_size * m_size - 1) As Integer
            Dim idx As Integer
            For Each s As String In vals
                If Not Integer.TryParse(s, tempArr(idx)) Then
                    Throw New IndexOutOfRangeException("Invalid value at position " & idx + 1)
                    Return
                End If
                If tempArr(idx) < 1 OrElse tempArr(idx) > vals.Length Then
                    Throw New IndexOutOfRangeException("Invalid value at position " & idx + 1 & vbCrLf & "only values in range 1 to " & vals.Length & " acceptable.")
                    Return
                End If
                idx += 1
            Next
            If tempArr(vals.Length - 1) <> vals.Length Then
                Throw New IndexOutOfRangeException("Empty value must be last in range!")
                Return
            End If
            cells = tempArr
            count = m_size * m_size
            m_board = New Board(cells)
        End Set
    End Property
    Public ReadOnly Property GameBoard As Board
        Get
            Return m_board
        End Get
    End Property

    Public Sub New(ByVal puzzleSize As Integer)
        m_size = puzzleSize
        InitPuzzle()
    End Sub
    Public Sub New(ByVal strConfig As String)
        StringConfig = strConfig
    End Sub

    Public Function Solveable() As Boolean
        emptyID = count
        Dim parity As Boolean = False
        For i As Integer = 0 To count - 2
            For j As Integer = i + 1 To count - 1
                If cells(j) < cells(i) Then parity = Not parity
            Next
            If cells(i) = count Then emptyID = i
        Next
        If (emptyID \ 4 Mod 2) = 0 Then parity = Not parity
        Return parity
    End Function

    Public Function Solve() As Solution
        Dim solver As New PuzzleSolver(m_size)
        Return solver.Solve(StringConfig)
    End Function

    Public Sub Reset()
        m_board = New Board(cells)
    End Sub

    Public Sub NewGame()
        InitPuzzle()
    End Sub

    Public Function TileSequence(ByVal sol As Solution) As List(Of Integer)
        Dim tmpBoard As Board = m_board
        Dim seq As New List(Of Integer)
        For Each m As Direction In sol.path
            seq.Add(tmpBoard.GetMoveTile(m))
            tmpBoard.Move(m)
        Next
        Return seq
    End Function
    Private Sub InitPuzzle()
        count = m_size * m_size
        cells = Enumerable.Range(1, count).ToArray
        Dim rnd As New Random
        For i As Integer = count - 2 To 0 Step -1
            Swap(rnd.Next(0, i), i)
        Next
        If Not Solveable() Then Swap(0, 1)
        m_board = New Board(cells)
    End Sub
    Private Sub Swap(ByVal a As Integer, ByVal b As Integer) 'swap 2 cells
        Dim n As Integer = cells(a)
        cells(a) = cells(b) : cells(b) = n
    End Sub

End Class
