﻿Public Class frmMain

    Private puzzleImage As Bitmap
    Private puzzle As Puzzle
    Private moveEmpty, bSolving As Boolean
    Private showDigits As Boolean = True
    Private imgPath As String
    Private tileImage() As Bitmap

#Region " Menu "
    Private Sub mnuOptions_Click(ByVal sender As Object, ByVal e As System.EventArgs) Handles mnuOptions.Click
        mnuSolve.Enabled = Not puzzle Is Nothing
    End Sub

    Private Sub mnuNewGame3_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuNewGame3.Click
        lblSolution.Text = ""
        lvSolution.Items.Clear()
        puzzle = New Puzzle(3)
        pbPuzzle.Image = New Bitmap(pbPuzzle.Width, pbPuzzle.Height)
        ResizeForm(False)
        DrawPuzzle()
    End Sub
    Private Sub mnuNewGame4_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuNewGame4.Click
        lblSolution.Text = ""
        lvSolution.Items.Clear()
        puzzle = New Puzzle(4)
        pbPuzzle.Image = New Bitmap(pbPuzzle.Width, pbPuzzle.Height)
        ResizeForm(False)
        DrawPuzzle()
    End Sub

    Private Sub mnuNewConfig_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuNewConfig.Click
        Dim s As String = InputBox("Enter tiles positions separatred by comma." & vbCrLf & "Tiles numbers starts from 1." & vbCrLf & _
                                   "Empty tile (with maximum number, 9 or 16) must be at last position", "Configure new puzzle")
        If Not String.IsNullOrEmpty(s) Then
            lblSolution.Text = ""
            lvSolution.Items.Clear()
            puzzle = New Puzzle(s)
            pbPuzzle.Image = New Bitmap(pbPuzzle.Width, pbPuzzle.Height)
            ResizeForm(False)
            DrawPuzzle()
        End If
    End Sub

    Private Sub mnuSolve_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuSolve.Click
        Dim filename As String = String.Format("puzzle_{0:d}x{1:d}.db", puzzle.BoardSize, puzzle.BoardSize)
        If System.IO.File.Exists(filename) Then
            lblSolution.Text = "Solving puzzle..."
        Else
            lblSolution.Text = "Generating lookup table for the first time." & vbCrLf & "This can takes few minutes..."
        End If
        bSolving = True
        lvSolution.Items.Clear()
        mnuOptions.Enabled = False
        ResizeForm(True)
        Me.Cursor = Cursors.WaitCursor
        Me.UseWaitCursor = True
        My.Application.DoEvents()
        puzzle.Reset()
        DrawPuzzle()
        Dim sol As Solution = puzzle.Solve
        If sol.path Is Nothing Then
            lblSolution.Text = "Path is empty - no moves"
            Return
        End If
        lblSolution.Text = "Optimal solution found in " & sol.path.Count & " moves."
        Dim tileSeq As List(Of Integer) = puzzle.TileSequence(sol)
        Dim n As Integer
        For Each p As Direction In sol.path
            lvSolution.Items.Add(p.ToString).SubItems.Add(tileSeq(n))
            n += 1
        Next
        mnuOptions.Enabled = True
        Me.Cursor = Cursors.Default
        Me.UseWaitCursor = False
        bSolving = False
    End Sub

    Private Sub mnuSettings_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuSettings.Click
        Using f As New frmSettings
            f.Init(imgPath, moveEmpty, showDigits)
            If f.ShowDialog = Windows.Forms.DialogResult.OK Then
                moveEmpty = f.optMoveEmpty.Checked
                showDigits = f.chkShowDigits.Checked
                If System.IO.File.Exists(f.txtImagePath.Text) Then
                    imgPath = f.txtImagePath.Text
                    puzzleImage = New Bitmap(imgPath)
                Else
                    imgPath = ""
                    puzzleImage = Nothing
                    showDigits = True
                End If
                DrawPuzzle()
            End If
        End Using
    End Sub

    Private Sub mnuExit_Click(ByVal sender As Object, ByVal e As System.EventArgs) Handles mnuExit.Click
        Me.Close()
    End Sub
#End Region

#Region " Drawing "
    Private Sub DrawPuzzle()
        SpliceImage(puzzleImage)
        pbPuzzle.Image = New Bitmap(puzzle.BoardSize * 100, puzzle.BoardSize * 100)
        Using g As Graphics = Graphics.FromImage(pbPuzzle.Image)
            For i As Integer = 0 To puzzle.BoardSize - 1
                For j As Integer = 0 To puzzle.BoardSize - 1
                    DrawCell(i, j)
                Next
            Next
        End Using
    End Sub

    Private Sub DrawCell(ByVal row As Integer, ByVal col As Integer)
        Dim s As String, n As Integer
        Dim sf As New StringFormat
        sf.Alignment = StringAlignment.Center
        sf.LineAlignment = StringAlignment.Center
        Using g As Graphics = Graphics.FromImage(pbPuzzle.Image)
            Dim rc As New Rectangle(col * 100 + 5, row * 100 + 5, 90, 90)
            If puzzle.GameBoard.Tiles(row, col) = puzzle.EmptyValue Then
                g.FillRectangle(Brushes.SkyBlue, rc)
            Else
                n = puzzle.GameBoard.Tiles(row, col)
                If puzzleImage Is Nothing Then
                    g.FillRectangle(Brushes.Silver, rc)
                Else
                    g.DrawImage(tileImage(n - 1), rc)
                End If
                g.DrawRectangle(Pens.Black, rc)
                If showDigits Then
                    s = n.ToString
                    g.DrawString(s, New Font("Tahoma", 18), Brushes.White, rc, sf)
                    rc.Offset(-2, -2)
                    g.DrawString(s, New Font("Tahoma", 18), Brushes.Black, rc, sf)
                End If
            End If
        End Using
    End Sub

    Private Sub SpliceImage(ByVal img As Bitmap)
        If img Is Nothing Then Return
        Dim imgNew As Bitmap = New Bitmap(img, puzzle.BoardSize * 90, puzzle.BoardSize * 90)
        ReDim tileImage(puzzle.BoardSize * puzzle.BoardSize - 1)
        Dim n As Integer
        For i As Integer = 0 To puzzle.BoardSize - 1
            For j As Integer = 0 To puzzle.BoardSize - 1
                tileImage(n) = New Bitmap(90, 90)
                Using g As Graphics = Graphics.FromImage(tileImage(n))
                    Dim rc As New Rectangle(j * 90, i * 90, 90, 90)
                    g.DrawImage(imgNew, 0, 0, rc, GraphicsUnit.Pixel)
                End Using
                n += 1
            Next
        Next
        imgNew.Dispose()
    End Sub

    Private Sub ResizeForm(ByVal bShowSolution As Boolean)
        If bShowSolution Then
            Me.Width = puzzle.BoardSize * 100 + lvSolution.Width + 30
        Else
            Me.Width = puzzle.BoardSize * 100 + 10
            Me.Height = puzzle.BoardSize * 100 + 60
            lvSolution.Left = Me.Width
            lvSolution.Height = puzzle.BoardSize * 100 - 40
            lblSolution.Top = lvSolution.Top + lvSolution.Height + 10
            lblSolution.Left = lvSolution.Left
        End If
        Me.Left = Screen.PrimaryScreen.WorkingArea.Width / 2 - Me.Width / 2
    End Sub

#End Region

    Private Sub pbPuzzle_MouseMove(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles pbPuzzle.MouseMove
        If bSolving Then Return
        pbPuzzle.Cursor = If(moveEmpty, GetCursorEmpty(e.Location), GetCursor(e.Location))
    End Sub

    Private Sub pbPuzzle_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles pbPuzzle.Click
        If bSolving Then Return
        Dim eRow As Integer = puzzle.GameBoard.EmptyTileY
        Dim ecol As Integer = puzzle.GameBoard.EmptyTileX
        Select Case pbPuzzle.Cursor
            Case Cursors.PanEast
                puzzle.GameBoard.Move(If(moveEmpty, Direction.RIGHT, Direction.LEFT))
            Case Cursors.PanWest
                puzzle.GameBoard.Move(If(moveEmpty, Direction.LEFT, Direction.RIGHT))
            Case Cursors.PanNorth
                puzzle.GameBoard.Move(If(moveEmpty, Direction.UP, Direction.DOWN))
            Case Cursors.PanSouth
                puzzle.GameBoard.Move(If(moveEmpty, Direction.DOWN, Direction.UP))
            Case Else
                Return
        End Select
        DrawCell(puzzle.GameBoard.EmptyTileY, puzzle.GameBoard.EmptyTileX)
        DrawCell(eRow, ecol)
        pbPuzzle.Refresh()
    End Sub

    Private Function GetCursor(ByVal pt As Point) As Cursor
        If puzzle Is Nothing Then Return Cursors.Default
        Dim row As Integer = Int(pt.Y / 100)
        Dim col As Integer = Int(pt.X / 100)
        If row < 0 OrElse row > puzzle.BoardSize - 1 Then Return Cursors.Default
        If col < 0 OrElse col > puzzle.BoardSize - 1 Then Return Cursors.Default
        Dim rc As New Rectangle(col * 100 + 5, row * 100 + 5, 90, 90)
        If Not rc.Contains(pt) Then Return Cursors.Default
        Dim emptyRow As Integer = puzzle.GameBoard.EmptyTileY
        Dim emptyCol As Integer = puzzle.GameBoard.EmptyTileX
        'curRow = row : curCol = col
        Dim cur As Cursor = Cursors.No
        If emptyCol = col Then
            If emptyRow - row = 1 Then
                cur = Cursors.PanSouth
            ElseIf row - emptyRow = 1 Then
                cur = Cursors.PanNorth
            End If
        ElseIf emptyRow = row Then
            If emptyCol - col = 1 Then
                cur = Cursors.PanEast
            ElseIf col - emptyCol = 1 Then
                cur = Cursors.PanWest
            End If
        End If
        Return cur
    End Function
    Private Function GetCursorEmpty(ByVal pt As Point) As Cursor
        If puzzle Is Nothing Then Return Cursors.Default
        Dim rc As New Rectangle(puzzle.GameBoard.EmptyTileX * 100 + 5, _
                                puzzle.GameBoard.EmptyTileY * 100 + 5, 90, 90)
        If Not rc.Contains(pt) Then Return Cursors.Default
        Dim ptCenter As New Point(rc.X + 45, rc.Y + 45)
        Dim dx As Integer = pt.X - rc.X
        Dim dy As Integer = pt.Y - rc.Y
        If dx < dy Then
            If dx < (90 - dy) * Math.Cos(Math.PI / 4) Then
                If puzzle.GameBoard.CanGoLeft Then Return Cursors.PanWest
            Else
                If puzzle.GameBoard.CanGoDown Then Return Cursors.PanSouth
            End If
        Else
            If dx < (90 - dy) * Math.Cos(Math.PI / 4) Then
                If puzzle.GameBoard.CanGoUp Then Return Cursors.PanNorth
            Else
                If puzzle.GameBoard.CanGoRight Then Return Cursors.PanEast
            End If
        End If
        Return Cursors.Default
    End Function

    Private Sub lvSolution_SelectedIndexChanged(ByVal sender As Object, ByVal e As System.EventArgs) Handles lvSolution.SelectedIndexChanged
        If bSolving Then Return
        If lvSolution.SelectedItems.Count = 0 Then Return
        puzzle.Reset()
        Dim idx As Integer = lvSolution.SelectedItems(0).Index
        If idx > -1 Then
            For i As Integer = 0 To idx
                Dim strMove As String = lvSolution.Items(i).SubItems(0).Text.ToUpper
                Select Case strMove
                    Case "LEFT"
                        puzzle.GameBoard.MoveLeft()
                    Case "RIGHT"
                        puzzle.GameBoard.MoveRight()
                    Case "UP"
                        puzzle.GameBoard.MoveUp()
                    Case "DOWN"
                        puzzle.GameBoard.MoveDown()
                End Select
            Next
        End If
        DrawPuzzle()
    End Sub

    Private Sub frmMain_Load(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles MyBase.Load
        lblSolution.Text = ""
        lvSolution.Items.Clear()
        puzzle = New Puzzle(4)
        pbPuzzle.Image = New Bitmap(pbPuzzle.Width, pbPuzzle.Height)
        ResizeForm(False)
        DrawPuzzle()
    End Sub
End Class
